/*
 * OdometricLocalizer.java
 * 
 * Copyright 2009 by RidgeSoft, LLC., PO Box 482, Pleasanton, CA  94566, U.S.A.
 * www.ridgesoft.com
 *
 * RidgeSoft grants you the right to use, modify, make derivative works and
 * redistribute this source file provided you do not remove this copyright notice.
 */

import com.ridgesoft.robotics.ShaftEncoder;

public class OdometricLocalizer extends Thread implements Localizer {
    private static final float PI = 3.14159f;
    private static final float TWO_PI = PI * 2.0f;

    private ShaftEncoder leftEncoder;
    private ShaftEncoder rightEncoder;
    private int previousLeftCounts;
    private int previousRightCounts;
    private float distancePerCount;
    private float radiansPerCount;
    private float x = 0.0f;
    private float y = 0.0f;
    private float heading = 0.0f;
    private int period;

    public OdometricLocalizer(ShaftEncoder leftEncoder,
            ShaftEncoder rightEncoder, float wheelDiameter, float trackWidth,
            int countsPerRevolution, int threadPriority, int period) {
        this.leftEncoder = leftEncoder;
        this.rightEncoder = rightEncoder;
        distancePerCount = (PI * wheelDiameter) / (float) countsPerRevolution;
        radiansPerCount = PI * (wheelDiameter / trackWidth)
                / countsPerRevolution;
        this.period = period;
        previousLeftCounts = leftEncoder.getCounts();
        previousRightCounts = rightEncoder.getCounts();
        setDaemon(true);
        setPriority(threadPriority);
        start();
    }

    public synchronized Pose getPose() {
        return new Pose(x, y, heading);
    }

    public void run() {
        try {
            long nextTime = System.currentTimeMillis();
            while (true) {
                // read encoders
                int leftCounts = leftEncoder.getCounts();
                int rightCounts = rightEncoder.getCounts();

                // calculate change in pose
                int deltaLeft = leftCounts - previousLeftCounts;
                int deltaRight = rightCounts - previousRightCounts;
                float deltaDistance = 0.5f * (float) (deltaLeft + deltaRight)
                        * distancePerCount;
                float deltaX = deltaDistance * (float) Math.cos(heading);
                float deltaY = deltaDistance * (float) Math.sin(heading);
                float deltaTheta = (float) (deltaRight - deltaLeft)
                        * radiansPerCount;

                // update pose
                synchronized (this) {
                    x += deltaX;
                    y += deltaY;
                    heading += deltaTheta;

                    // limit theta to -Pi <= theta < Pi
                    if (heading > PI)
                        heading -= TWO_PI;
                    else if (heading <= -PI)
                        heading += TWO_PI;
                }

                previousLeftCounts = leftCounts;
                previousRightCounts = rightCounts;

                nextTime += period;
                Thread.sleep(nextTime - System.currentTimeMillis());
            }
        }
        catch (Throwable t) {
            t.printStackTrace();
        }
    }
}